import pytest
import numpy as np
import warnings
import pmesh
from mpi4py import MPI
from hymd.input_parser import Config, _find_unique_names
from hymd.thermostat import csvr_thermostat
from hymd.field import domain_decomposition
from hymd.file_io import distribute_input


@pytest.mark.mpi()
def test_thermostat_coupling_groups(molecules_with_solvent):
    R = 0.00831446261815324  # kJ/mol K, gas constant

    class RandomMock:
        ind = 0

        def __init__(self, x):
            self.x = x

        def __call__(self, *args):
            self.ind += 1
            return self.x[self.ind - 1]

    def T_from_K(K, N):
        return 2 * K / (3 * R * N)

    def K_from_V(V, comm=MPI.COMM_WORLD, m=72.0):
        return 0.5 * m * comm.allreduce(np.sum(V**2), MPI.SUM)

    def K_from_T(T, N):
        return 3 * R * T * N / 2

    (indices_global, positions_global, molecules_global, velocities_global,
     bonds_global, names_global, types_global) = molecules_with_solvent

    n_particles = len(indices_global)
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(
        n_steps=0, time_step=0.032958582578275, box_size=box_size,
        tau=0.925852989520023, mesh_size=[2, 2, 2], sigma=0.5, kappa=0.05,
        n_particles=len(indices_global), target_temperature=310.0,
        thermostat_work=0.0, mass=72.0
    )
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action='ignore', category=np.VisibleDeprecationWarning,
            message=r'Creating an ndarray from ragged nested sequences'
        )
        pm = pmesh.ParticleMesh(config.mesh_size, BoxSize=config.box_size,
                                dtype='f8', comm=comm)

    # Test stub acting like a hdf5 file for distribute_input
    in_file = {'molecules': molecules_global, 'indices': indices_global}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size,  config.n_particles, max_molecule_size=6,
        comm=comm,
    )
    positions = positions_global[rank_range]
    molecules = molecules_global[rank_range]
    indices = indices_global[rank_range]
    velocities = velocities_global[rank_range]
    bonds = bonds_global[rank_range]
    types = types_global[rank_range]
    names = names_global[rank_range]

    dd = domain_decomposition(
        positions, pm, indices, velocities, names, types, molecules=molecules,
        bonds=bonds, verbose=2, comm=comm,
    )
    positions, indices, velocities, names, types, bonds, molecules = dd

    # Regression testing to ensure the underlying system didn't change.
    kinetic_energy = 168.45555165866017
    temperature = 300.1562358223504
    kinetic_energy_species = [55.69986940118755, 59.56037533981311,
                              7.495327677224677, 45.699979240434836]
    temperature_species = [319.0073556406238, 318.3762937488012,
                           600.9871410378249, 244.28640571781278]

    total_kinetic_energy = K_from_V(velocities, comm=comm)
    total_temperature = T_from_K(total_kinetic_energy, n_particles)
    assert total_kinetic_energy == pytest.approx(kinetic_energy, abs=1e-12)
    assert total_temperature == pytest.approx(temperature, abs=1e-12)

    for t, T, K in zip(
        ("A", "B", "C", "D"), temperature_species, kinetic_energy_species
    ):
        n_particles_species = comm.allreduce(
            len(np.where(names == np.string_(t))[0]), MPI.SUM
        )
        K_species = K_from_V(velocities[names == np.string_(t)], comm=comm)
        T_species = T_from_K(K_species, n_particles_species)
        assert K == pytest.approx(K_species, abs=1e-13)
        assert T == pytest.approx(T_species, abs=1e-13)

    config = _find_unique_names(config, names, comm=comm)
    velocities_copy = velocities.copy()

    # create np.random.Generator just to pass as argument
    prng = np.random.default_rng()

    # Hijack the _random_gaussian and _random_chi_squared interal functions in
    # thermostat to control the exact numbers generated for testing.
    random_chi_squared = RandomMock([125.4595634810623])
    random_gaussian = RandomMock([0.5579657512081987])
    csvr_thermostat(
        velocities_copy, names, config, prng, comm=comm,
        random_gaussian=random_gaussian, random_chi_squared=random_chi_squared,
        remove_center_of_mass_momentum=False,
    )
    kinetic_energy_new = K_from_V(velocities_copy, comm=comm)
    assert kinetic_energy_new == pytest.approx(171.25339969021243, abs=1e-13)
    assert config.thermostat_work == pytest.approx(2.797848031552252, abs=1e-13)

    config.thermostat_work = 0.0
    config.thermostat_coupling_groups = [
        ["A"],
        ["B"],
        ["C"],
        ["D"],
    ]
    velocities_copy = velocities.copy()
    random_chi_squared = RandomMock([35.43824087713971, 27.57022975113815,
                                     2.024725328228174, 36.50208472031436])
    random_gaussian = RandomMock([-1.752320325907187, 1.099694957420625,
                                  0.6113448515745533, -0.7183266831611322])
    csvr_thermostat(
        velocities_copy, names, config, prng, comm=comm,
        random_gaussian=random_gaussian, random_chi_squared=random_chi_squared,
        remove_center_of_mass_momentum=False,
    )
    assert config.thermostat_work == pytest.approx(-21.670791766960217, abs=1e-13)  # noqa: E501

    config.thermostat_work = 0.0
    config.thermostat_coupling_groups = [
        ["A", "B", "C"],
        ["D"],
    ]
    velocities_copy = velocities.copy()
    random_chi_squared = RandomMock([97.07130218590895, 33.06359496718739])
    random_gaussian = RandomMock([0.1661408606772054, -0.06216797747541603])
    csvr_thermostat(
        velocities_copy, names, config, prng, comm=comm,
        random_gaussian=random_gaussian, random_chi_squared=random_chi_squared,
        remove_center_of_mass_momentum=False,
    )

    inds_ABC = np.where(
        np.logical_or.reduce(
            (names == np.string_("A"), names == np.string_("B"),
             names == np.string_("C"))
        )
    )
    K_group_ABC = K_from_V(velocities_copy[inds_ABC], comm=comm)
    K_group_D = K_from_V(velocities_copy[names == np.string_("D")], comm=comm)

    # assert K_group_ABC == pytest.approx(123.6090634948376, abs=1e-13)
    # assert K_group_D == pytest.approx(45.41754237593906, abs=1e-13)
    # assert config.thermostat_work == pytest.approx(0.5710542121164457,
    #                                                abs=1e-13)



"""Integrates equations of motion.

Implements the Velocity Verlet integrator. The *reversible reference system
propagator algorithm* (rRESPA) integrator uses the Velocity Verlet functions
inside the MD loop in the main module.
"""


def integrate_velocity(velocities, accelerations, time_step):
    """Velocity update step of the Velocity Verlet integration algorithm.

    Computes the velocity update step of the Velocity Verlet algorithm. The
    :code:`velocities` argument is **not** changed in place.

    Parameters
    ----------
    velocities : (N, D) numpy.ndarray
        Array of velocities of :code:`N` particles in :code:`D` dimensions.
    accelerations : (N, D) numpy.ndarray
        Array of accelerations of :code:`N` particles in :code:`D` dimensions.
    time_step : float
        The time step used in the integration.

    Notes
    -----
    The Velocity Verlet algorithm contains two steps, first the velocties are
    integrated one half step forward in time by applying the forces from the
    previous step

    .. math:: \\mathbf{v}_{\\text{new}} = \\mathbf{v} + \\frac{\\Delta t}{2m}
              \\mathbf{f}

    before the positions are moved a full step using the updated half-step
    velocities.

    See Also
    --------
    integrate_position : The position update step of the Velocity Verlet
                         algorithm.
    """
    return velocities + 0.5 * time_step * accelerations


def integrate_position(positions, velocities, time_step):
    """Position update step of the Velocity Verlet integration algorithm.

    Computes the position update step of the Velocity Verlet algorithm. The
    :code:`positions` argument is **not** changed in place.

    Parameters
    ----------
    positions : (N, D) numpy.ndarray
        Array of positions of :code:`N` particles in :code:`D` dimensions.
    velocities : (N, D) numpy.ndarray
        Array of velocities of :code:`N` particles in :code:`D` dimensions.
    time_step : float
        The time step used in the integration.

    Notes
    -----
    The Velocity Verlet algorithm contains two steps, first the velocties are
    integrated one half step forward in time by applying the forces from the
    previous step, then the positions are updated a full step using the updated
    velocities

    .. math:: \\mathbf{x}_{\\text{new}} = \\mathbf{x} + \\Delta t \\mathbf{v}

    See Also
    --------
    integrate_velocity : The velocity update step of the Velocity Verlet
                         algorithm.
    """
    return positions + time_step * velocities



"""Scales or otherwise modifies the particle velocities during simulation to
simulate coupling to an external heat bath with temperature T₀.
"""
import numpy as np
import logging
from mpi4py import MPI
from typing import Callable
from .logger import Logger
from .input_parser import Config


def cancel_com_momentum(velocities, config, comm=MPI.COMM_WORLD):
    com_velocity = comm.allreduce(np.sum(velocities[...], axis=0), MPI.SUM)
    velocities[...] = velocities[...] - com_velocity / config.n_particles
    return velocities


def generate_initial_velocities(velocities, config, prng, comm=MPI.COMM_WORLD):
    kT_start = config.gas_constant * config.start_temperature
    n_particles_ = velocities.shape[0]
    velocities[...] = prng.normal(
        loc=0, scale=kT_start / config.mass, size=(n_particles_, 3)
    )
    velocities = cancel_com_momentum(velocities, config, comm=comm)
    kinetic_energy = comm.allreduce(
        0.5 * config.mass * np.sum(velocities**2), MPI.SUM
    )
    start_kinetic_energy_target = (
        1.5 * config.gas_constant * config.n_particles * config.start_temperature
    )
    factor = np.sqrt(1.5 * config.n_particles * kT_start / kinetic_energy)
    velocities[...] = velocities[...] * factor
    kinetic_energy = comm.allreduce(
        0.5 * config.mass * np.sum(velocities**2), MPI.SUM
    )
    Logger.rank0.log(
        logging.INFO,
        (
            f"Initialized {config.n_particles} velocities, target kinetic "
            f"energy: {start_kinetic_energy_target}, actual kinetic energy "
            f"generated: {kinetic_energy}"
        ),
    )
    return velocities


def _random_gaussian(prng: np.random.Generator) -> float:
    """Draw a single random number from the standard normal distribution
    Generate a number from the Gaussian distribution centered at zero with unit
    standard deviation, :math:`N(0, 1)`.

    Parameters
    ----------
    prng : np.random.Generator
        Numpy object that provides a stream of random bits

    Returns
    -------
    float
        A random number drawn from :math:`N(0, 1)`.
    """
    return prng.normal()


def _random_chi_squared(prng: np.random.Generator, M: int) -> float:
    """Draw the sum of :code:`M` squared normally distributed values

    The value is generated by the Gamma distribution, in lieu of generating
    :code:`M` Gaussian distributed numbers and summing their squares.

    Parameters
    ----------
    prng : np.random.Generator
        Numpy object that provides a stream of random bits
    M : int
        Number of standard normally distributed numbers in the sum.

    Returns
    -------
    float
        The sum of :code:`M` squared normally distributed values centered at
        zero with unit standard deviation.

    Notes
    -----
    The sum of the squares of k independent standard normal random variables is
    distributed according to a :math:`\\chi^2` distribution.

    .. math::

        \\chi^2_1 + \\chi_2^2 + \\chi_3^2 + \\dots + \\chi_M^2
        \\sim
        \\sigma^2\\chi^2(k)

    This is a special case of the :math:`\\Gamma` distribution,
    :math:`\\Gamma(k/2, 2)`, and may be generated by

    .. math::

        \\chi^2(k) \\sim 2 \\Gamma(k/2, 2)

    References
    ----------
    Knuth, D.E. 1981, Seminumerical Algorithms, 2nd ed., vol. 2 of The Art of
    Computer Programming (Reading, MA: Addison-Wesley), pp. 120ff.
    J. H. Ahrens and U. Dieter, Computing 12 (1974), 223-246.
    """
    return prng.chisquare(M)


def csvr_thermostat(
    velocity: np.ndarray,
    names: np.ndarray,
    config: Config,
    prng: np.random.Generator,
    comm: MPI.Intracomm = MPI.COMM_WORLD,
    random_gaussian: Callable[[np.random.Generator], float] = _random_gaussian,
    random_chi_squared: Callable[
        [np.random.Generator, int, float], float
    ] = _random_chi_squared,
    remove_center_of_mass_momentum: bool = True,
) -> np.ndarray:
    """Canonical sampling through velocity rescaling thermostat

    Implements the CSVR thermostat. Rescales the system kinetic energy by a
    stochastically chosen factor to keep the temperature constant. Requires
    communcation of the kinetic energies calculated locally for each MPI rank.
    The random numbers sampled through :code:`random_gaussian` and
    :code:`random_chi_squared` are broadcast from the root rank to the other
    ranks to ensure the scaling is performed with the same stochasticity for
    all particles in the full system.

    The velocities are cleaned of center of mass momentum before the thermostat
    is applied, and the center of mass momentum is subsequently reapplied. This
    is performed for each thermostat coupling group, i.e. the center of mass
    momenta of each *group* is separately removed and reapplied after
    thermostatting.

    The implementation here is based on the derivation presented in the 2008
    Comput. Phys. Commun paper, not in the original 2007 J. Chem. Phys. paper.

    Parameters
    ----------
    velocity : (N, D) numpy.ndarray
        Array of velocities of N particles in D dimensions.
    names : (N,) numpy.ndarray
        Array of particle names.
    config : Config
        Configuration dataclass containing simulation metadata and parameters.
    prng : np.random.Generator
        Numpy object that provides a stream of random bits
    comm : MPI.Intracomm, optional
        MPI communicator to use for rank commuication. Defaults to
        MPI.COMM_WORLD.
    random_gaussian : callable, optional
        Function for generating standard normally distributed numbers.
    random_chi_squared : callable, optional
        Function for generating :math:`\\chi^2`-distributed numbers
    remove_center_of_mass_momentum : bool, optional
        If True, the center of mass of each coupling group is removed before
        the thermostat is applied. The center of mass momenta are added back
        after the kinetic energy rescaling is complete.

    See Also
    --------
    _random_gaussian :
        Used to sample Gaussian-distributed numbers.
    _random_chi_squared :
        Used to sample :math:`\\chi^2`-distributed numbers.
    hymd.input_parser.Config :
        Configuration dataclass handler.

    References
    ----------
    G. Bussi, D. Donadio, and M. Parrinello, J. Chem. Phys. 126, 014101 (2007).
    G. Bussi and M. Parrinello, Comput. Phys. Commun. 179, 26-29, (2008).
    """
    if not any(config.thermostat_coupling_groups):
        config.thermostat_coupling_groups = [config.unique_names.copy()]
    for i, group in enumerate(config.thermostat_coupling_groups):
        ind = np.where(
            np.logical_or.reduce(list(names == np.string_(t) for t in group))
        )
        group_n_particles = comm.allreduce(len(ind[0]), MPI.SUM)

        # Clean velocities of center of mass momentum
        if remove_center_of_mass_momentum and group_n_particles > 1:
            com_velocity = comm.allreduce(np.sum(velocity[ind], axis=0))
            velocity_clean = velocity[ind] - com_velocity / group_n_particles
            K = comm.allreduce(0.5 * config.mass * np.sum(velocity_clean[...] ** 2))
        else:
            K = comm.allreduce(0.5 * config.mass * np.sum(velocity[...] ** 2))
        K_target = (
            1.5 * config.gas_constant * group_n_particles * config.target_temperature
        )
        N_f = 3 * group_n_particles
        c = np.exp(-(config.time_step * config.respa_inner) / config.tau)

        # Random numbers are identical across ranks because all ranks have
        # the same prng after the initialization of the velocities
        R = random_gaussian(prng)
        SNf = random_chi_squared(prng, N_f - 1)

        alpha2 = (
            c
            + (1 - c) * (SNf + R**2) * K_target / (N_f * K)
            + 2 * R * np.sqrt(c * (1 - c) * K_target / (N_f * K))
        )
        dK = K * (alpha2 - 1)
        alpha = np.sqrt(alpha2)

        if remove_center_of_mass_momentum and group_n_particles > 1:
            # Assign velocities and reapply the previously removed center
            # of mass momentum removed
            velocity_clean *= alpha
            velocity[ind] = velocity_clean + com_velocity / group_n_particles
        else:
            velocity *= alpha
        config.thermostat_work += dK



import logging
import numpy as np
from mpi4py import MPI
from .logger import Logger
import sympy
from .field import comp_laplacian


def comp_pressure(
    phi,
    phi_q,
    psi,
    hamiltonian,
    velocities,
    config,
    phi_fourier,
    phi_laplacian,
    phi_transfer,
    positions,
    bond_pr,
    angle_pr,
    comm=MPI.COMM_WORLD,
):
    """
    Computes total internal pressure of the system.
    Kinetic pressure is trivially calculated from the kinetic energy.
    Already computed bond and angle pressure terms are inserted into the
    total internal pressure.
    The field pressure equation is implemented.

    Parameters
    ----------
    phi : list[pmesh.pm.RealField], (M,)
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type
        :code:`M`. Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    velocities : (N, D) numpy.ndarray
        Array of velocities of N particles in D dimensions.
    config : Config
        Configuration dataclass containing simulation metadata and parameters.
    phi_fourier : list[pmesh.pm.ComplexField], (M,)
        Pmesh :code:`ComplexField` objects containing discretized particle
        number density values in reciprocal space on the computational grid;
        one for each particle type. Pre-allocated, but empty; any values in
        this field are discarded Changed in-place. Local for each MPI rank--the
        full computaional grid is represented by the collective fields of all
        MPI ranks.
    phi_laplacian : list[pmesh.pm.RealField], (M, 3)
        Like phi, but containing the laplacian of particle number densities.
    phi_transfer : list[pmesh.pm.ComplexField], (3,)
        Like phi_fourier, used as an intermediary to perform FFT operations
        to obtain the gradient or laplacian of particle number densities.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    bond_pr : (3,) numpy.ndarray
        Total bond pressure due all two-particle bonds.
    angle_pr : (3,) numpy.ndarray
        Total angle pressure due all three-particle bonds
    comm : MPI.Intracomm, optional
        MPI communicator to use for rank commuication. Defaults to
        MPI.COMM_WORLD.

    Returns
    -------
    pressure : (18,) numpy.ndarray
        Pressure contributions from various energy terms.
        0: due to kinetic energy
        1-5: due to field energy
        6-8: due to two-particle bonded terms
        9-11: due to three-particle bonded terms (called angle terms)
        12-14: due to four-particle bonded terms (called dihedral terms)
        (defaults to 0 currently. Yet to be implemented)
        15-17: total pressure in x,y,z directions.
    """
    rank = comm.Get_rank()
    size = comm.Get_size()

    V = np.prod(config.box_size)
    n_mesh__cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh__cells

    w_0 = hamiltonian.w_0(phi) * volume_per_cell
    w_elec = 0.0
    if (config.coulombtype == "PIC_Spectral_GPE") or (
        config.coulombtype == "PIC_Spectral"
    ):
        w_elec = hamiltonian.w_elec([phi_q, psi]) * volume_per_cell
    w = w_0 + w_elec

    # Kinetic term
    kinetic_energy = 0.5 * config.mass * np.sum(velocities**2)
    p_kin = 2 / (3 * V) * kinetic_energy

    # Term 1
    p0 = -1 / V * np.sum(w)

    # Term 2
    V_bar = np.array([hamiltonian.V_bar[k]([phi, psi]) for k in range(config.n_types)])

    p1 = np.sum((volume_per_cell / V) * V_bar * phi)

    # Term 3
    comp_laplacian(
        phi_fourier,
        phi_transfer,
        phi_laplacian,
        hamiltonian,
        config,
    )

    p2 = np.sum(
        volume_per_cell
        / V
        * config.sigma**2
        * np.repeat(V_bar[:, np.newaxis, :, :, :], 3, axis=1)
        * phi_laplacian,
        axis=(0, 2, 3, 4),
    )

    # Bonded force term: linking 2 particles
    p_bond = bond_pr / V

    # Angle force term: linking 3 particles
    p_angle = angle_pr / V

    # TODO: Dihedral angle force term: linking 4 atoms
    p_dihedral = np.zeros(3)

    # Add formal parameter dihedral_forces as: comp_pressure(..., dihedral_forces)
    # Define dictionary:
    # forces = {
    #          'x': dihedral_forces[:,0],
    #          'y': dihedral_forces[:,1],
    #          'z': dihedral_forces[:,2]
    #           }
    # Compute the pressure due to dihedrals as:
    # p_dihedral = {
    #          'x': np.sum( np.multiply(forces['x'],positions[:,0]) )*(1/V),
    #          'y': np.sum( np.multiply(forces['y'],positions[:,1]) )*(1/V),
    #          'z': np.sum( np.multiply(forces['z'],positions[:,2]) )*(1/V)
    #              }

    # Total pressure in x, y, z
    p_tot = p_kin + p0 + p1 + p2 + p_bond + p_angle + p_dihedral

    return_value = np.array(
        [
            p_kin,
            p0,
            p1,
            p2[0],
            p2[1],
            p2[2],
            p_bond[0],
            p_bond[1],
            p_bond[2],
            p_angle[0],
            p_angle[1],
            p_angle[2],
            p_dihedral[0],
            p_dihedral[1],
            p_dihedral[2],
            p_tot[0],
            p_tot[1],
            p_tot[2],
        ]
    )
    return_value = comm.allreduce(return_value, MPI.SUM)

    return return_value



"""Parses command line arguments to HyMD
"""
from argparse import ArgumentParser
import os
import sys
import numpy as np
import atexit
import cProfile
import logging
import pstats
import tomli
from .logger import Logger, print_header
from .input_parser import read_config_toml, parse_config_toml


def configure_runtime(args_in, comm):
    """Parse command line arguments and configuration file

    Parameters
    ----------
    args_in : list
        List with arguments to be processed.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    Returns
    -------
    args : argparse.Namespace
        Namespace containing command line arguments.
    config : hymd.input_parser.Config
        Parsed configuration object.
    """
    ap = ArgumentParser()

    ap.add_argument(
        "-v",
        "--verbose",
        default=1,
        type=int,
        nargs="?",
        help="Increase logging verbosity",
    )
    ap.add_argument(
        "--profile",
        default=False,
        action="store_true",
        help="Profile program execution with cProfile",
    )
    ap.add_argument(
        "--disable-field",
        default=False,
        action="store_true",
        help="Disable field forces",
    )
    ap.add_argument(
        "--disable-bonds",
        default=False,
        action="store_true",
        help="Disable two-particle bond forces",
    )
    ap.add_argument(
        "--disable-angle-bonds",
        default=False,
        action="store_true",
        help="Disable three-particle angle bond forces",
    )
    ap.add_argument(
        "--disable-dihedrals",
        default=False,
        action="store_true",
        help="Disable four-particle dihedral forces",
    )
    ap.add_argument(
        "--disable-dipole",
        default=False,
        action="store_true",
        help="Disable BB dipole calculation",
    )
    ap.add_argument(
        "--double-precision",
        default=False,
        action="store_true",
        help="Use double precision positions/velocities",
    )
    ap.add_argument(
        "--double-output",
        default=False,
        action="store_true",
        help="Use double precision in output h5md",
    )
    ap.add_argument(
        "--dump-per-particle",
        default=False,
        action="store_true",
        help="Log energy values per particle, not total",
    )
    ap.add_argument(
        "--force-output",
        default=False,
        action="store_true",
        help="Dump forces to h5md output",
    )
    ap.add_argument(
        "--velocity-output",
        default=False,
        action="store_true",
        help="Dump velocities to h5md output",
    )
    ap.add_argument(
        "--disable-mpio",
        default=False,
        action="store_true",
        help=("Avoid using h5py-mpi, potentially decreasing IO " "performance"),
    )
    ap.add_argument(
        "--destdir", default=".", help="Write output to specified directory"
    )
    ap.add_argument(
        "--seed",
        default=None,
        type=int,
        help="Set the numpy random generator seed for every rank",
    )
    ap.add_argument(
        "--logfile",
        default="sim.log",
        help="Redirect event logging to specified file",
    )
    ap.add_argument(
        "-p", "--topol", default=None, help="Gmx-like topology file in toml format"
    )
    ap.add_argument("config", help="Config .py or .toml input configuration script")
    ap.add_argument("input", help="input.hdf5")
    args = ap.parse_args(args_in)

    if comm.Get_rank() == 0:
        os.makedirs(args.destdir, exist_ok=True)
    comm.barrier()

    # Safely define seeds
    seeds = None
    if comm.Get_rank() == 0:
        if args.seed is not None:
            ss = np.random.SeedSequence(args.seed)
        else:
            ss = np.random.SeedSequence()
        seeds = ss.spawn(comm.Get_size())

    seeds = comm.bcast(seeds, root=0)

    # Setup a PRNG for each rank
    prng = np.random.default_rng(seeds[comm.Get_rank()])

    # Setup logger
    Logger.setup(
        default_level=logging.INFO,
        log_file=f"{args.destdir}/{args.logfile}",
        verbose=args.verbose,
    )

    # print header info
    Logger.rank0.log(logging.INFO, print_header())

    if args.profile:
        prof_file_name = "cpu.txt-%05d-of-%05d" % (comm.rank, comm.size)
        output_file = open(os.path.join(args.destdir, prof_file_name), "w")
        pr = cProfile.Profile()

        def profile_atexit():
            pr.disable()
            # Dump results:
            # - for binary dump
            prof_file_bin = "cpu.prof-%05d-of-%05d" % (comm.rank, comm.size)
            pr.dump_stats(os.path.join(args.destdir, prof_file_bin))
            stats = pstats.Stats(pr, stream=output_file)
            stats.sort_stats("time").print_stats()
            output_file.close()

        # TODO: if we have a main function then we can properly do set up and
        # teardown without using atexit.
        atexit.register(profile_atexit)

        pr.enable()

    try:
        Logger.rank0.log(
            logging.INFO,
            f"Attempting to parse config file {args.config} as " ".toml",
        )

        if args.topol is not None:
            topol = tomli.loads(read_config_toml(args.topol))
            # Check if we have single "itp" files and add their keys to topol
            if os.path.dirname(args.topol) == "":
                args.topol = "./" + args.topol
            if "include" in topol["system"]:
                for file in topol["system"]["include"]:
                    path = f"{os.path.dirname(args.topol)}/{file}"
                    itps = tomli.loads(read_config_toml(path))
                    for mol, itp in itps.items():
                        topol[mol] = itp
        else:
            topol = None
        toml_config = read_config_toml(args.config)
        config = parse_config_toml(
            toml_config, file_path=os.path.abspath(args.config), comm=comm
        )
        Logger.rank0.log(
            logging.INFO, f"Successfully parsed {args.config} as .toml file"
        )
        config.command_line_full = " ".join(sys.argv)
        Logger.rank0.log(logging.INFO, str(config))
    except ValueError as ve:
        raise ValueError(
            f"Unable to parse configuration file {args.config}"
            f"\n\ntoml parse traceback:" + repr(ve)
        )
    return args, config, prng, topol



"""Implements the stochastic cell rescaling (SCR) barostat.
This is a first-order barostat that samples the correct
volume fluctuations through a suitable noise term.
Scales the box and particle positions by a factor (see functions
`isotropic` and `semiisotropic` for details) during simulation to
simulate coupling to an external pressure bath set at a
target pressure.

The box and particle positions are scaled in the L and N directions according
to the nature of the barostat (see functions `isotropic` and `semiisotropic`
below by an amount :math:`\\alpha`.

The updated system information is passed on to the pmesh objects.

References
----------
Mattia Bernetti and Giovanni Bussi , "Pressure control using stochastic
cell rescaling", J. Chem. Phys. 153, 114107 (2020)
"""
import numpy as np
from mpi4py import MPI
from dataclasses import dataclass
from typing import Union
from .pressure import comp_pressure
from .field import initialize_pm


@dataclass
class Target_pressure:
    P_L: Union[bool, float]
    P_N: Union[bool, float]


def isotropic(
    pmesh,
    pm_stuff,
    phi,
    phi_q,
    psi,
    hamiltonian,
    positions,
    velocities,
    config,
    phi_fft,
    phi_laplacian,
    phi_transfer,
    bond_pr,
    angle_pr,
    step,
    prng,
    comm=MPI.COMM_WORLD,
):
    """
    It calculates the scaling factor according to

    .. math::

        \\log \\alpha' = - \\frac{n_b\,dt\,\\beta}{\\tau_p}(P^t - P) + \\sqrt{\\frac{2n_b^2k_BT\\beta\,dt}{\mathcal{V}\\tau_p}}dW

    .. math::

        \\alpha = \\exp{\\frac{1}{3}\\log \\alpha'}

    where :math:`dt` is the outer rRESPA time-step, :math:`n_b` is the frequency
    of barostat calls, :math:`\\tau_p` is the pressure coupling time constant,
    :math:`\\beta` is the isothermal compressibility, :math:`P_{L,N}^t` and
    :math:`P_{L,N}` is the target and instantaneous internal pressure in the
    lateral (L) and normal (N) directions respectively. Convention: Cartesian
    z-direction is considered normal.
    """
    rank = comm.Get_rank()
    beta = 7.6 * 10 ** (-4)  # bar^(-1) #isothermal compressibility of water
    change = False

    if np.mod(step, config.n_b) == 0:

        R = prng.normal()

        change = True
        # compute pressure
        pressure = comp_pressure(
            phi,
            phi_q,
            psi,
            hamiltonian,
            velocities,
            config,
            phi_fft,
            phi_laplacian,
            phi_transfer,
            positions,
            bond_pr,
            angle_pr,
            comm=comm,
        )

        # Total pressure across all ranks
        P = np.average(pressure[-3:-1])  # kJ/(mol nm^3)
        P = P * 16.61  # bar

        V = np.prod(config.box_size)
        noise_term = (
            np.sqrt(
                2.0
                * config.n_b
                * config.gas_constant
                * config.target_temperature
                * beta
                * config.time_step
                * config.n_b
                / (V * config.tau_p)
            )
            * R
        )
        log_alpha = (
            -config.n_b
            * config.time_step
            * beta
            / config.tau_p
            * (config.target_pressure.P_L - P)
        )
        log_alpha = log_alpha + noise_term
        alpha = np.exp(log_alpha / 3.0)

        config.box_size *= alpha

        # position coordinates scaling
        positions *= alpha

        # pmesh re-initialize
        pm_stuff = initialize_pm(pmesh, config, comm)
    return (pm_stuff, False)


def semiisotropic(
    pmesh,
    pm_stuff,
    phi,
    phi_q,
    psi,
    hamiltonian,
    positions,
    velocities,
    config,
    phi_fft,
    phi_laplacian,
    phi_transfer,
    bond_pr,
    angle_pr,
    step,
    prng,
    comm=MPI.COMM_WORLD,
):
    """
    It calculates the scaling factor according to

    .. math::

        \\log \\alpha'_{L} = - \\frac{2n_b\,dt\,\\beta}{3\\tau_p}(P_L^t - P_L - \\frac{\\gamma}{L_z}) + \\sqrt{\\frac{4n_b^2k_BT\\beta\,dt}{3\mathcal{V}\\tau_p}}dW_L

    .. math::

        \\alpha_L = \\exp{\\frac{1}{2}\\log \\alpha'_L}

    .. math::

        \\log \\alpha'_{N} = - \\frac{n_b\,dt\,\\beta}{3\\tau_p}(P_N^t - P_N) + \\sqrt{\\frac{2n_b^2k_BT\\beta\,dt}{3\mathcal{V}\\tau_p}}dW_N

    .. math::

        \\alpha_L = \\log \\alpha'_N

    where :math:`dt` is the outer rRESPA time-step, :math:`n_b` is the frequency
    of barostat calls, :math:`\\tau_p` is the pressure coupling time constant,
    :math:`\\beta` is the isothermal compressibility, :math:`P_{L,N}^t` and
    :math:`P_{L,N}` is the target and instantaneous internal pressure in the
    lateral (L) and normal (N) directions respectively, :math:`\\gamma` is the
    surface tension. Convention: Cartesian z-direction is considered normal.
    """
    rank = comm.Get_rank()
    beta = 7.6 * 10 ** (-4)  # isothermal compressibility of water
    change = False
    if np.mod(step, config.n_b) == 0:
        Rxy = prng.normal()
        Rz = prng.normal()
        change = True
        # compute pressure
        pressure = comp_pressure(
            phi,
            phi_q,
            psi,
            hamiltonian,
            velocities,
            config,
            phi_fft,
            phi_laplacian,
            phi_transfer,
            positions,
            bond_pr,
            angle_pr,
            comm=comm,
        )

        # Total pressure across all ranks
        # L: Lateral; N: Normal
        [PL, PN] = [0, 0]
        PL = (pressure[-3] + pressure[-2]) / 2  # kJ/(mol nm^3)
        PN = pressure[-1]  # kJ/(mol nm^3)
        PL = PL * 16.61  # bar
        PN = PN * 16.61  # bar
        alphaL = 1.0
        alphaN = 1.0
        config.surface_tension = config.box_size[2] / 2 * (PN - PL)  # bar nm

        if config.target_pressure.P_L:
            V = np.prod(config.box_size)
            noise_term = (
                np.sqrt(
                    4.0
                    * config.n_b
                    * config.gas_constant
                    * config.target_temperature
                    * beta
                    * config.time_step
                    * config.n_b
                    / (3 * V * config.tau_p)
                )
                * Rxy
            )
            log_alpha = (
                -2.0
                * config.n_b
                * config.time_step
                * beta
                / (3 * config.tau_p)
                * (
                    config.target_pressure.P_L
                    - PL
                    - config.surface_tension / config.box_size[2]
                )
            )
            log_alpha = log_alpha + noise_term
            alpha = np.exp(
                log_alpha / 2.0
            )  # not 100% sure about this factor 2, have to check it out <<< TODO

            config.box_size[0:2] *= alpha

            positions[:][0:2] *= alpha

        if config.target_pressure.P_N:
            V = np.prod(config.box_size)
            noise_term = (
                np.sqrt(
                    2.0
                    * config.n_b
                    * config.gas_constant
                    * config.target_temperature
                    * beta
                    * config.time_step
                    * config.n_b
                    / (3 * V * config.tau_p)
                )
                * Rz
            )
            log_alpha = (
                -config.n_b
                * config.time_step
                * beta
                / (3 * config.tau_p)
                * (config.target_pressure.P_N - PN)
            )
            log_alpha = log_alpha + noise_term
            alpha = np.exp(
                log_alpha / 1.0
            )  # not 100% sure about this factor 2, have to check it out <<< TODO

            config.box_size[2] *= alpha

            positions[:][2] *= alpha

        # pmesh re-initialize
        pm_stuff = initialize_pm(pmesh, config, comm)
    return (pm_stuff, change)



import pytest
from mpi4py import MPI
import numpy as np
import collections
import itertools
import h5py
from hymd.file_io import (
    OutDataset, setup_time_dependent_element, store_static, store_data,
    distribute_input
)
from hymd.input_parser import Config, parse_config_toml, _setup_type_to_name_map
from hymd.force import prepare_bonds
from hymd.barostat import Target_pressure

def test_OutDataset(config_toml, tmp_path):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    out = OutDataset(tmp_path, config)
    assert hasattr(out, "config")
    assert hasattr(out, "disable_mpio")
    assert not out.disable_mpio
    assert hasattr(out, "float_dtype")
    assert out.float_dtype == "float32"
    assert hasattr(out, "file")
    assert isinstance(out.file, h5py.File)
    assert out.file # check if open

    out.flush()
    out.close_file()
    assert not out.file # check if closed

    out = OutDataset(tmp_path, config, double_out=True)
    assert out.float_dtype == "float64"
    out.close_file()

    out = OutDataset(tmp_path, config, disable_mpio=True)
    assert out.disable_mpio
    assert isinstance(out.file, h5py.File)
    assert out.file # check if open
    out.close_file()
    assert not out.file


def test_setup_time_dependent_element(config_toml, tmp_path):
    _, config_toml_str = config_toml
    config = parse_config_toml(config_toml_str)

    out = OutDataset(tmp_path, config)
    test_group = out.file.create_group("/test")

    (
        group, step, time, value
    ) = setup_time_dependent_element(
        "position",
        test_group,
        1,
        (config.n_particles, 3),
        "float32",
        units="nm"
    )

    assert isinstance(group, h5py.Group)
    assert group.name == "/test/position"

    assert isinstance(step, h5py.Dataset)
    assert isinstance(time, h5py.Dataset)
    assert time.attrs["unit"] == "ps"
    assert isinstance(value, h5py.Dataset)
    assert value.attrs["unit"] == "nm"

    out.close_file()


@pytest.mark.mpi()
def test_distribute_input():
    # create fake indices and molecules
    indices = np.arange(0, 10000)
    molecules = np.zeros_like(indices)
    molecules[400:450] = 1
    molecules[450:] = np.arange(2, 9552)
    assert indices.shape == molecules.shape

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    in_file = {'indices': indices}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size,  None, comm=comm
    )
    assert not molecules_flag
    gather_range = comm.allgather(rank_range)
    gather_range = list(itertools.chain.from_iterable(gather_range))
    assert np.array_equal(gather_range, indices)

    in_file = {'indices': indices, 'molecules': molecules}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size, indices.shape[0], max_molecule_size=1000, comm=comm
    )
    assert molecules_flag
    gather_range = comm.allgather(rank_range)
    gather_range = list(itertools.chain.from_iterable(gather_range))
    assert np.array_equal(gather_range, indices)


@pytest.mark.mpi()
def test_store_static(molecules_with_solvent, tmp_path):
    indices, positions, molecules, velocities, bonds, names, types = molecules_with_solvent
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(n_steps=0, n_print=1, time_step=0.03, box_size=box_size,
                    mesh_size=[5, 5, 5], sigma=0.5, kappa=0.05,
                    n_particles=len(indices), 
                    target_pressure=Target_pressure(P_L=None,P_N=None))
    Bond = collections.namedtuple(
        "Bond", ["atom_1", "atom_2", "equilibrium", "strength"]
    )
    cbonds = (
        Bond("A", "A", 0.27, 10000),
        Bond("A", "B", 0.27, 10000),
        Bond("A", "C", 0.27, 10000),
        Bond("B", "B", 0.27, 10000),
        Bond("B", "C", 0.27, 10000),
    )
    config.bonds = cbonds
    config = _setup_type_to_name_map(config, names, types)

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    test_dir = comm.bcast(tmp_path, root=0)
    out = OutDataset(test_dir, config, comm=comm)

    # Test stub acting like a hdf5 file for distribute_input
    in_file = {'molecules': molecules, 'indices': indices}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size,  config.n_particles, 6, comm=comm
    )
    positions_ = positions[rank_range]
    molecules_ = molecules[rank_range]
    indices_ = indices[rank_range]
    velocities_ = velocities[rank_range]
    bonds_ = bonds[rank_range]
    types_ = types[rank_range]
    names_ = names[rank_range]

    bonds_prep = prepare_bonds(
        molecules_, names_, bonds_, indices_, config
    )
    (
        # two-particle bonds
        bonds_2_atom1, bonds_2_atom2, bonds_2_equilibrium,
        bonds_2_stength,
        # three-particle bonds
        bonds_3_atom1, bonds_3_atom2, bonds_3_atom3,
        bonds_3_equilibrium, bonds_3_stength,
        # four-particle bondsl
        bonds_4_atom1, bonds_4_atom2, bonds_4_atom3, bonds_4_atom4,
        bonds_4_coeff, bonds_4_type, bonds_4_last,
    ) = bonds_prep


    store_static(
        out, rank_range, names_, types_, indices_, config, bonds_2_atom1,
        bonds_2_atom2, molecules=molecules_
    )

    groups = ['connectivity', 'h5md', 'observables', 'parameters', 'particles']
    assert all((k in out.file.keys()) for k in groups)
    groups = ['box', 'mass', 'position', 'species']
    assert all((k in out.file['particles/all']) for k in groups)
    groups = [
        'angle_energy', 'angular_momentum', 'bond_energy', 
        'dihedral_energy', 'field_energy', 'kinetic_energy', 
        'potential_energy', 'temperature', 'thermostat_work', 
        'torque', 'total_energy', 'total_momentum'
    ]
    assert all((k in out.file['observables']) for k in groups)
    assert 'vmd_structure' in out.file['parameters'].keys()

    out.close_file()


@pytest.mark.mpi()
def test_store_data(molecules_with_solvent, tmp_path):
    indices, positions, molecules, velocities, bonds, names, types = molecules_with_solvent
    box_size = np.array([10, 10, 10], dtype=np.float64)
    config = Config(n_steps=100, n_print=1, time_step=0.03, box_size=box_size,
                    mesh_size=[5, 5, 5], sigma=0.5, kappa=0.05,
                    n_particles=len(indices), mass=72.,
                    target_pressure=Target_pressure(P_L=None,P_N=None))
    Bond = collections.namedtuple(
        "Bond", ["atom_1", "atom_2", "equilibrium", "strength"]
    )
    cbonds = (
        Bond("A", "A", 0.27, 10000),
        Bond("A", "B", 0.27, 10000),
        Bond("A", "C", 0.27, 10000),
        Bond("B", "B", 0.27, 10000),
        Bond("B", "C", 0.27, 10000),
    )
    config.bonds = cbonds
    config = _setup_type_to_name_map(config, names, types)

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    test_dir = comm.bcast(tmp_path, root=0)
    out = OutDataset(test_dir, config, comm=comm)

    # Test stub acting like a hdf5 file for distribute_input
    in_file = {'molecules': molecules, 'indices': indices}
    rank_range, molecules_flag = distribute_input(
        in_file, rank, size,  config.n_particles, 6, comm=comm
    )
    positions_ = positions[rank_range]
    molecules_ = molecules[rank_range]
    indices_ = indices[rank_range]
    velocities_ = velocities[rank_range]
    bonds_ = bonds[rank_range]
    types_ = types[rank_range]
    names_ = names[rank_range]

    bonds_prep = prepare_bonds(
        molecules_, names_, bonds_, indices_, config
    )
    (
        # two-particle bonds
        bonds_2_atom1, bonds_2_atom2, bonds_2_equilibrium,
        bonds_2_stength,
        # three-particle bonds
        bonds_3_atom1, bonds_3_atom2, bonds_3_atom3,
        bonds_3_equilibrium, bonds_3_stength,
        # four-particle bondsl
        bonds_4_atom1, bonds_4_atom2, bonds_4_atom3, bonds_4_atom4,
        bonds_4_coeff, bonds_4_type, bonds_4_last,
    ) = bonds_prep


    # call it to prepare OutDataset
    store_static(
        out, rank_range, names_, types_, indices_, config, bonds_2_atom1,
        bonds_2_atom2, molecules=molecules_
    )

    forces = np.copy(positions_)

    store_data(
        out, 0, 0, indices_, positions_, velocities_, forces,
        box_size, 300., 1., 1., 2., 3., 4., 5., 6., 0.02, config
    )

    outdataset_step = [
        out.positions_step,
        out.total_energy_step,
        out.potential_energy_step,
        out.kinetc_energy_step,
        out.bond_energy_step,
        out.angle_energy_step,
        out.dihedral_energy_step,
        out.field_energy_step,
        out.total_momentum_step,
        out.angular_momentum_step,
        out.torque_step,
        out.temperature_step,
        out.thermostat_work_step,
    ]

    outdataset_time = [
        out.positions_time,
        out.total_energy_time,
        out.potential_energy_time,
        out.kinetc_energy_time,
        out.bond_energy_time,
        out.angle_energy_time,
        out.dihedral_energy_time,
        out.field_energy_time,
        out.total_momentum_time,
        out.angular_momentum_time,
        out.torque_time,
        out.temperature_time,
        out.thermostat_work_time,
    ]

    for dset in outdataset_step:
        assert isinstance(dset, h5py.Dataset)
        print(dset)
        assert dset[0] == 0

    for dset in outdataset_time:
        assert isinstance(dset, h5py.Dataset)
        assert dset[0] == 0

    assert out.potential_energy[0] == pytest.approx(20.)
    assert out.kinetc_energy[0] == pytest.approx(1.)
    assert out.temperature[0] == pytest.approx(300.)
    assert out.pressure[0] == pytest.approx(1.)
    assert out.bond_energy[0] == pytest.approx(2.)
    assert out.angle_energy[0] == pytest.approx(3.)
    assert out.dihedral_energy[0] == pytest.approx(4.)
    assert out.field_energy[0] == pytest.approx(5.)

    out.close_file()



import pytest
import logging
import numpy as np
import os
from mpi4py import MPI
from argparse import Namespace
from hymd.configure_runtime import configure_runtime
from hymd.input_parser import Config

def test_configure_runtime(h5py_molecules_file, config_toml,
                           change_tmp_dir, caplog):
    caplog.set_level(logging.ERROR)
    comm = MPI.COMM_WORLD
    config_toml_file, _ = config_toml

    basearg = [config_toml_file, h5py_molecules_file]

    parsed, config, prng, topol = configure_runtime(basearg, comm)
    assert isinstance(parsed, Namespace)
    assert isinstance(config, Config)
    assert isinstance(prng, np.random.Generator)
    assert parsed.destdir == "."
    assert parsed.logfile == "sim.log"

    parsed, _, _, _ = configure_runtime(["-v", "2"]+basearg, comm)
    assert parsed.verbose == 2

    parsed, _, _, _ = configure_runtime(["--disable-field"]+basearg, comm)
    assert parsed.disable_field

    parsed, _, _, _ = configure_runtime(["--disable-bonds"]+basearg, comm)
    assert parsed.disable_bonds

    parsed, _, _, _ = configure_runtime(["--disable-angle-bonds"]+basearg, comm)
    assert parsed.disable_angle_bonds

    parsed, _, _, _ = configure_runtime(["--disable-dihedrals"]+basearg, comm)
    assert parsed.disable_dihedrals

    parsed, _, _, _ = configure_runtime(["--disable-dipole"]+basearg, comm)
    assert parsed.disable_dipole

    parsed, _, _, _ = configure_runtime(["--double-precision"]+basearg, comm)
    assert parsed.double_precision

    parsed, _, _, _ = configure_runtime(["--double-output"]+basearg, comm)
    assert parsed.double_output

    parsed, _, _, _ = configure_runtime(["--dump-per-particle"]+basearg, comm)
    assert parsed.dump_per_particle

    parsed, _, _, _ = configure_runtime(["--force-output"]+basearg, comm)
    assert parsed.force_output

    parsed, _, _, _ = configure_runtime(["--velocity-output"]+basearg, comm)
    assert parsed.velocity_output

    parsed, _, _, _ = configure_runtime(["--disable-mpio"]+basearg, comm)
    assert parsed.disable_mpio

    parsed, _, _, _ = configure_runtime(["--destdir", "testdir"]+basearg, comm)
    assert parsed.destdir == "testdir"

    parsed, _, _, _ = configure_runtime(["--seed", "54321"]+basearg, comm)
    assert parsed.seed == 54321

    parsed, _, _, _ = configure_runtime(["--logfile", "test.log"]+basearg, comm)
    assert parsed.logfile == "test.log"



import setuptools  # noqa: F401, https://stackoverflow.com/a/55358607/4179419
from numpy.distutils.core import setup, Extension


def find_version(path):
    import re
    s = open(path, 'rt').read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              s, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Version not found")


with open('README.md', 'r') as in_file:
    readme = in_file.read()

force_kernels = Extension(
    name="force_kernels",
    sources=[
        "hymd/compute_bond_forces.f90",
        "hymd/compute_bond_forces__double.f90",
        "hymd/compute_angle_forces.f90",
        "hymd/compute_angle_forces__double.f90",
        "hymd/compute_dihedral_forces.f90",
        "hymd/compute_dihedral_forces__double.f90",
        "hymd/dipole_reconstruction.f90",
        "hymd/dipole_reconstruction__double.f90",
    ]
)

setup(
    name="hymd",
    author="Morten Ledum",
    author_email="morten.ledum@gmail.com",
    url="https://github.com/Cascella-Group-UiO/HyMD",
    description="Massively parallel hybrid particle-field MD",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="LGPLv3",
    packages=["hymd"],
    version=find_version("hymd/version.py"),
    ext_modules=[force_kernels],
    setup_requires=[
        "cython",
        "numpy<1.24",
        "mpi4py",
    ],
    install_requires=[
        "cython",
        "h5py",
        "mpi4py",
        "mpsort",
        "networkx",
        "numpy<1.24",
        "pfft-python",
        "pmesh",
        "sympy",
        "tomli",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",  # noqa: E501
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Fortran",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)



import pytest
import numpy as np
from hymd.integrator import integrate_velocity, integrate_position


def test_integrator_velocity_zero_force(three_atoms):
    _, _, _, _, _, velocities, CONF = three_atoms
    old_velocities = velocities.copy()
    velocities = integrate_velocity(velocities, np.zeros_like(velocities),
                                    CONF['dt'])
    assert np.allclose(old_velocities, velocities, atol=1e-13)


def test_integrator_velocity(three_atoms):
    _, _, _, _, _, velocities, CONF = three_atoms
    accelerations = np.array([
        [5.309962614030253,  -5.068488339109831,  0.828090598458989],
        [-3.868819942226709, -9.720221218265905,  4.036326041075613],
        [9.728114214745183,  -8.189855111379165, -3.075260427419437]
    ], dtype=np.float64)
    velocities = integrate_velocity(velocities, accelerations, CONF['dt'])
    expected_velocities = np.array([
        [0.3004100792669800, -0.452528545222956, -0.288323467370162],
        [-0.153011424114385,  0.313703442307804,  0.420088246650352],
        [0.4402634172106540,  0.077338401336559,  0.041682352532849]
    ], dtype=np.float64)
    assert np.allclose(expected_velocities, velocities, atol=1e-13)


def test_integrator_position_zero_velocity_zero_force(three_atoms):
    _, _, _, _, positions, velocities, CONF = three_atoms
    old_positions = positions.copy()
    positions = integrate_position(positions, np.zeros_like(velocities),
                                   CONF['dt'])
    assert np.allclose(old_positions, positions, atol=1e-13)


def test_integrator_position_zero_force(three_atoms):
    _, _, _, _, positions, velocities, CONF = three_atoms
    positions = integrate_position(positions, velocities, CONF['dt'])
    expected_positions = np.array([
        [3.341937188559520, 3.053303418939576, 3.648148235458336],
        [4.254612887747975, 3.315211337414966, 4.720346322206166],
        [2.170585066308332, 2.215015584448489, 4.184927411402398]
    ], dtype=np.float64)
    assert np.allclose(expected_positions, positions, atol=1e-13)


def test_integrator_position(three_atoms):
    _, _, _, _, positions, velocities, CONF = three_atoms
    positions = integrate_position(positions, velocities, CONF['dt'])
    expected_positions = np.array([
        [3.341937188559520, 3.053303418939576, 3.648148235458336],
        [4.254612887747975, 3.315211337414966, 4.720346322206166],
        [2.170585066308332, 2.215015584448489, 4.184927411402398]
    ], dtype=np.float64)
    assert np.allclose(expected_positions, positions, atol=1e-13)


def test_integrator_velocity_verlet_constant_force(three_atoms):
    """With a constant force F₀ = (α, β, γ), the velocity obeys the recurrence
    relation

        v       =   v      +  Δt  a
          i+1         i             0

    with solution

        v       =   v      +  i Δt a
          i           0              0

    The position obeys the recurrence relation
                                 ╭     ╮       1    2
        p       =   p      +  Δt │ v   │   +  ─── Δt  a
          i+1         i          ╰   i ╯       2        0
                                 ╭                 ╮       1    2
                =   p      +  Δt │ v   +  i Δt a   │   +  ─── Δt  a
                      i          ╰   0           0 ╯       2        0

    with solution
                               i Δt  ╭                      ╮
        p       =   p      +  ────── │ a   Δt i   +   2 v   │
          i           0         2    ╰   0                0 ╯
    """
    _, _, _, _, positions, velocities, CONF = three_atoms
    initial_positions = positions.copy()
    initial_velocities = velocities.copy()
    dt = CONF['dt']
    accelerations = np.array([
        [-7.539285072366672,  0.968464568407796,  0.598639645160741],
        [9.1962508228909070,  1.505356899114361,  3.601585312042335],
        [8.8869056807446400, -0.972763307999720, -7.335044988262844]
    ], dtype=np.float64)

    for step in range(1, 10):
        # Velocity Verlet steps
        velocities = integrate_velocity(velocities, accelerations, dt)
        positions = integrate_position(positions, velocities, dt)
        velocities = integrate_velocity(velocities, accelerations, dt)

        for i in range(3):
            for d in range(3):
                v = velocities[i, d]
                a = accelerations[i, d]
                v_0 = initial_velocities[i, d]
                v_expected = v_0 + step * dt * a
                assert v == pytest.approx(v_expected, abs=1e-13)

                p = positions[i, d]
                p_0 = initial_positions[i, d]
                p_expected = p_0 + 0.5 * step * dt * (a * dt * step + 2 * v_0)
                assert p == pytest.approx(p_expected, abs=1e-13)


def test_integrator_velocity_verlet_linear_force(three_atoms):
    """With a time-linear force F(t) = (α t, β t, γ t) the velocity obeys the
    relation
                             α Δt ╭            ╮
       v       =   v     +  ───── │ 2 t  +  Δt │
         i+1         0        2   ╰            ╯

    with solution
                             1      2   2
       v       =   v     +  ─── α Δt  i               (1)
         i           0       2

    The position recurrence relation takes the form
                               ╭         α    2  2 ╮    1      3
       p       =   p     +  Δt │ v    + ─── Δt  i  │ + ─── α Δt  i
         i+1         0         ╰   0     2         ╯    2

    with solution
                             1      3   ╭  2      ╮
       p       =   p     +  ─── α Δt  i │ i  -  1 │ +  Δt i v
         i           0       6          ╰         ╯           0
    """
    _, _, _, _, positions, velocities, CONF = three_atoms
    initial_positions = positions.copy()
    initial_velocities = velocities.copy()
    dt = CONF['dt']
    acceleration_constants = np.array([
        [-7.539285072366672,  0.968464568407796,  0.598639645160741],
        [9.1962508228909070,  1.505356899114361,  3.601585312042335],
        [8.8869056807446400, -0.972763307999720, -7.335044988262844]
    ], dtype=np.float64)

    # Initial forces at t=0 are zero
    accelerations = np.zeros_like(acceleration_constants)
    for step in range(1, 10):
        # Velocity Verlet steps
        velocities = integrate_velocity(velocities, accelerations, dt)
        positions = integrate_position(positions, velocities, dt)

        # Update forces
        accelerations = step * dt * acceleration_constants
        velocities = integrate_velocity(velocities, accelerations, dt)

        for i in range(3):
            for d in range(3):
                v = velocities[i, d]
                a_0 = acceleration_constants[i, d]
                v_0 = initial_velocities[i, d]
                v_expected = v_0 + 0.5 * a_0 * dt**2 * step**2
                assert v == pytest.approx(v_expected, abs=1e-13)

                p = positions[i, d]
                p_0 = initial_positions[i, d]
                p_expected = (p_0 + (step**2 - 1) * a_0 * dt**3 * step / 6.0 +
                              dt * step * v_0)
                assert p == pytest.approx(p_expected, abs=1e-13)


def test_integrator_respa_velocity_verlet_equivalent(three_atoms):
    _, _, _, _, positions_vv, velocities_vv, CONF = three_atoms
    positions_respa = positions_vv.copy()
    velocities_respa = velocities_vv.copy()
    Dt = CONF['dt']
    M = 1  # Inner RESPA steps per long Δt steps
    dt = Dt / M

    def acceleration_long(positions):
        equilibrium = np.array([
            [5.5643712591568950,  0.884023577514768,  9.170162581620790],
            [2.8788490727172800,  3.400482961264622, -1.520671282680709],
            [-1.487026411226955, -6.859348424240592, -2.549294472000168]
        ], dtype=np.float64)
        return (positions - equilibrium)**2

    def acceleration_short(positions):
        return - positions

    a_l_vv = acceleration_long(positions_vv)
    a_s_vv = acceleration_short(positions_vv)
    a_l_respa = acceleration_long(positions_respa)
    a_s_respa = acceleration_short(positions_respa)

    for step in range(1, 10):
        # RESPA steps
        velocities_respa = integrate_velocity(velocities_respa, a_l_respa, Dt)

        for s in range(M):
            velocities_respa = integrate_velocity(velocities_respa, a_s_respa,
                                                  dt)
            positions_respa = integrate_position(positions_respa,
                                                 velocities_respa, dt)

            a_s_respa = acceleration_short(positions_respa)
            velocities_respa = integrate_velocity(velocities_respa, a_s_respa,
                                                  dt)

        a_l_respa = acceleration_long(positions_respa)
        velocities_respa = integrate_velocity(velocities_respa, a_l_respa, Dt)

        # Velocity Verlet steps
        velocities_vv = integrate_velocity(velocities_vv, a_l_vv + a_s_vv, Dt)
        positions_vv = integrate_position(positions_vv, velocities_vv, Dt)
        a_l_vv = acceleration_long(positions_vv)
        a_s_vv = acceleration_short(positions_vv)
        velocities_vv = integrate_velocity(velocities_vv, a_l_vv + a_s_vv, Dt)

        assert np.allclose(velocities_vv, velocities_respa, atol=1e-13)
        assert np.allclose(positions_vv, positions_respa, atol=1e-13)


def test_integrator_respa_constant_force(three_atoms):
    """With a constant force F₀ = (α, β, γ), the velocity obeys the recurrence
    relation (dt = Δt / M, with M the number of inner RESPA steps per long Δt
    step)

        v       =   v      +  Δt a      +  M dt a
          i+1         i            L              S

    with solution
                                ╭                       ╮
        v       =   v      +  i │ Δt a      +  M dt a   │
          i           0         ╰      L              S ╯

    The position obeys the recurrence relation                  2
                        M dt  ╭                         ╮     dt  aS (M - 1) M
        p    =  p   +  ────── │ 2 v   + Δt a   + dt a   │ +  ──────────────────
         i+1      i      2    ╰     i        L        S ╯            4
    """
    _, _, _, _, initial_positions, initial_velocities, CONF = three_atoms
    Dt = CONF['dt']
    acceleration_long = np.array([
        [1.6851177297118020,  2.849556517742162,  1.109363734104484],
        [-3.129093587905032,  1.363258438204340,  3.621192799979525],
        [0.2028711865869220, -3.176012574568037, -2.357588753443123]
    ], dtype=np.float64)
    acceleration_short = np.array([
        [-6.162908853722353,  2.993015003002288,  9.462957452628899],
        [2.2878276475508310, -4.867859096429122, -8.123570401997174],
        [-7.826918369186258, -3.199027970628050, -6.067654493641959]
    ], dtype=np.float64)

    for respa in (1, 2, 3, 4, 5, 10, 20):
        M = respa
        dt = Dt / M

        velocities = initial_velocities.copy()
        positions = initial_positions.copy()

        for step in range(1, 10):
            v_o = velocities.copy()
            p_o = positions.copy()

            # RESPA steps
            velocities = integrate_velocity(velocities, acceleration_long, Dt)

            for s in range(M):
                velocities = integrate_velocity(velocities, acceleration_short,
                                                dt)
                positions = integrate_position(positions, velocities, dt)

                # Calculate forces_short here
                velocities = integrate_velocity(velocities, acceleration_short,
                                                dt)

            # Calculate forces_long here
            velocities = integrate_velocity(velocities, acceleration_long, Dt)

            for i in range(3):
                for d in range(3):
                    v = velocities[i, d]
                    v_0 = initial_velocities[i, d]
                    p = positions[i, d]
                    # p_0 = initial_positions[i, d]
                    a_s = acceleration_short[i, d]
                    a_l = acceleration_long[i, d]
                    v_expected = v_0 + step * (Dt * a_l + M * dt * a_s)
                    assert v == pytest.approx(v_expected, abs=1e-13)

                    p_expected = p_o[i, d] + (
                        dt * M * v_o[i, d] +
                        dt * M / 2 * Dt * a_l +
                        dt * M**2 / 2 * dt * a_s
                    )
                    assert p == pytest.approx(p_expected, abs=1e-13)
